from typing import Any
from service_forge.workflow.registry.sf_base_model import SfBaseModel
from service_forge.workflow.node import Node
from service_forge.workflow.port import Port

class ParsePaginationResponseNodeType(SfBaseModel):
    items: list[Any]
    page: int
    page_size: int
    total: int

class ParsePaginationResponseNode(Node):
    DEFAULT_INPUT_PORTS = [
        Port("response", ParsePaginationResponseNodeType),
        Port("assert_not_empty", bool),
    ]

    DEFAULT_OUTPUT_PORTS = [
        Port("items", list[Any]),
        Port("page", int),
        Port("page_size", int),
        Port("total", int),
        Port("first_item", Any),
    ]

    def __init__(self, name: str):
        super().__init__(name)

    async def _run(self, response: ParsePaginationResponseNodeType, assert_not_empty: bool) -> None:
        self.activate_output_edges('items', response.items)
        self.activate_output_edges('page', response.page)
        self.activate_output_edges('page_size', response.page_size)
        self.activate_output_edges('total', response.total)

        if assert_not_empty:
            if len(response.items) == 0:
                raise ValueError("Pagination response items is empty")

        self.activate_output_edges('first_item', response.items[0] if len(response.items) > 0 else None)


from __future__ import annotations
from sqlalchemy.orm import DeclarativeBase
from service_forge.utils.register import Register
from service_forge.db.trace_mixin import register_trace_events

register_trace_events(DeclarativeBase)

class SfDbBase(DeclarativeBase):
    __abstract__ = True
    
    def __init_subclass__(cls) -> None:
        if cls.__name__ not in ["SfDbBase"]:
            sf_dbbase_register.register(cls.__name__, cls)
        return super().__init_subclass__()

sf_dbbase_register = Register[SfDbBase]()



import os
from service_forge.workflow.node import Node
from service_forge.workflow.port import Port
from model.node.reply import Reply
from model.http.query_model import QueryModel
from service_forge.llm import Model
import litellm

class ReplyNode(Node):
    DEFAULT_INPUT_PORTS = [
        Port("user_input", QueryModel),
        Port("prompt", str),
        Port("temperature", float),
    ]

    DEFAULT_OUTPUT_PORTS = [
        Port("reply", Reply),
    ]

    def __init__(self, name: str):
        super().__init__(name)

    async def _run(self, prompt: str, user_input: QueryModel, temperature: float) -> None:
        if os.path.exists(prompt):
            with open(prompt, "r") as f:
                prompt = f.read()
        else:
            prompt = prompt

        print("Querying LLM...")
        response = self.llm.completion(
            input=user_input.user_input,
            system_prompt=prompt,
            model=Model.DEEPSEEK_V3_250324,
            temperature=temperature,
        )

        result = Reply(reply=response.choices[0].message.content)

        self.activate_output_edges(self.get_output_port_by_name('reply'), result)


import os
import httpx
from fastapi import APIRouter, HTTPException, Query, BackgroundTasks, Request
from loguru import logger
from typing import Optional
from datetime import datetime

from service_forge.model.feedback import FeedbackCreate, FeedbackResponse, FeedbackListResponse
from service_forge.storage.feedback_storage import feedback_storage
from service_forge.db.database import create_database_manager
from service_forge.current_service import get_service

router = APIRouter(prefix="/sdk/feedback", tags=["feedback"])

def _get_feedback_api_base():
    _service = get_service()
    if not _service:
        return None
    if not _service.config.feedback:
        return None
    return _service.config.feedback.api_url

def _get_feedback_api_timeout():
    _service = get_service()
    if not _service:
        return 10
    if not _service.config.feedback:
        return 10
    return _service.config.feedback.api_timeout

async def fetch_trace_info_from_trace(trace_id: str) -> dict:
    """
    从 trace API 获取 trace 的完整信息（service_name 和 workflow_name）

    Args:
        trace_id: trace ID

    Returns:
        包含 service_name 和 workflow_name 的字典
    """
    trace_service_url = _get_feedback_api_base()
    if not trace_service_url:
        logger.debug("未配置 trace service URL，跳过从 trace 获取信息")
        return {}

    # 使用新的 trace info API 端点，无需提供服务名
    trace_info_url = f"{trace_service_url}/sdk/trace/info"
    timeout = _get_feedback_api_timeout()

    try:
        async with httpx.AsyncClient(timeout=timeout) as client:
            response = await client.post(
                trace_info_url,
                json={"trace_id": trace_id}
            )
            response.raise_for_status()
            data = response.json()

            result = {}
            if data.get("service_name"):
                result["service_name"] = data.get("service_name")
            if data.get("workflow_name"):
                result["workflow_name"] = data.get("workflow_name")

            if result:
                logger.info(f"从 trace API 获取信息成功: trace_id={trace_id}, service_name={result.get('service_name')}, workflow_name={result.get('workflow_name')}")
                return result
            else:
                logger.warning(f"trace API 返回中未找到有效信息: trace_id={trace_id}")
                return {}

    except httpx.TimeoutException:
        logger.warning(f"trace API 请求超时: {trace_info_url}")
        return {}
    except httpx.ConnectError as e:
        logger.warning(f"trace API 连接失败: {trace_info_url} - {e}")
        return {}
    except httpx.HTTPStatusError as e:
        logger.error(f"trace API 请求失败: status={e.response.status_code}")
        return {}
    except Exception as e:
        logger.error(f"从 trace API 获取信息异常: {type(e).__name__}: {e}")
        return {}

async def fetch_workflow_name_from_trace(trace_id: str) -> Optional[str]:
    """
    从 trace info API 获取 workflow_name

    Args:
        trace_id: 追踪ID

    Returns:
        workflow_name 或 None
    """
    trace_service_url = _get_feedback_api_base()
    if not trace_service_url:
        logger.debug("未配置 trace service URL，跳过从 trace 获取 workflow_name")
        return None

    # 使用新的 trace info API 端点，无需提供服务名
    trace_info_url = f"{trace_service_url}/sdk/trace/info"
    timeout = _get_feedback_api_timeout()

    try:
        async with httpx.AsyncClient(timeout=timeout) as client:
            response = await client.post(
                trace_info_url,
                json={"trace_id": trace_id}
            )
            response.raise_for_status()
            data = response.json()

            # 从响应中提取 workflow_name
            workflow_name = data.get("workflow_name")
            if workflow_name:
                logger.info(f"从 trace API 获取 workflow_name 成功: trace_id={trace_id}, workflow_name={workflow_name}")
                return workflow_name
            else:
                logger.warning(f"trace API 返回中未找到 workflow_name: trace_id={trace_id}")
                return None

    except httpx.TimeoutException:
        logger.warning(f"从 trace API 获取 workflow_name 超时: {trace_info_url}")
        return None
    except httpx.ConnectError as e:
        logger.warning(f"从 trace API 获取 workflow_name 连接失败: {trace_info_url} - {e}")
        return None
    except httpx.HTTPStatusError as e:
        logger.warning(f"从 trace API 获取 workflow_name 请求失败: status={e.response.status_code}")
        return None
    except Exception as e:
        logger.warning(f"从 trace API 获取 workflow_name 异常: {type(e).__name__}: {e}")
        return None

async def forward_feedback_to_api(feedback_data: dict, user_id: Optional[str] = None, auth_token: Optional[str] = None, is_forwarded: bool = False):
    """
    将反馈数据转发到外部 API

    Args:
        feedback_data: 反馈数据字典
        user_id: 用户ID（从 request.state 获取）
        auth_token: 认证 token（从 request.state 获取）
        is_forwarded: 是否已经被转发过（用于防止递归转发）
    """
    # 如果已经被转发过，不再转发（防止递归转发）
    if is_forwarded:
        logger.debug(f"反馈已被转发过，跳过重复转发: feedback_id={feedback_data.get('feedback_id')}")
        return

    feedback_service_base_url = _get_feedback_api_base()
    if not feedback_service_base_url:
        logger.debug("未配置转发 API URL，跳过转发")
        return

    auto_send_url = f"{feedback_service_base_url}/sdk/feedback"
    timeout = _get_feedback_api_timeout()
    headers = {
        "Content-Type": "application/json",
        # 添加标记表示这是转发的反馈，防止递归转发
        "X-Feedback-Forwarded": "true",
    }

    # 从 request state 中添加用户认证信息
    if user_id:
        headers["X-User-ID"] = user_id
    if auth_token:
        headers["X-User-Token"] = auth_token

    try:
        # 处理 datetime 对象，转换为 ISO 格式字符串
        serializable_data = feedback_data.copy()
        if 'created_at' in serializable_data and serializable_data['created_at']:
            serializable_data['created_at'] = serializable_data['created_at'].isoformat()

        logger.info(f"\n{'='*60}")
        logger.info(f"转发反馈到外部 API")
        logger.info(f"URL: {auto_send_url}")
        logger.info(f"Method: POST")
        logger.info(f"Headers: {headers}")
        logger.info(f"Payload: {serializable_data}")
        logger.info(f"{'='*60}\n")

        async with httpx.AsyncClient(timeout=timeout, follow_redirects=True) as client:
            response = await client.post(
                auto_send_url,
                json=serializable_data,
                headers=headers
            )
            response.raise_for_status()
            logger.info(f"✓ 反馈转发成功: feedback_id={feedback_data.get('feedback_id')}, URL={auto_send_url}, status={response.status_code}")
    except httpx.TimeoutException:
        logger.warning(f"✗ 反馈转发超时: {auto_send_url}")
    except httpx.ConnectError as e:
        logger.warning(f"✗ 反馈转发连接失败: {auto_send_url} - {e}")
    except httpx.HTTPStatusError as e:
        logger.error(f"✗ 反馈转发失败: URL={auto_send_url}, status={e.response.status_code}, detail={e.response.text}")
    except Exception as e:
        logger.error(f"✗ 反馈转发异常: URL={auto_send_url}, {type(e).__name__}: {e}")

async def fetch_feedbacks_from_external_api(trace_id: str) -> list[dict]:
    """
    从外部反馈服务 API 获取反馈数据

    Args:
        trace_id: 追踪ID

    Returns:
        反馈数据列表
    """
    feedback_service_base_url = _get_feedback_api_base()
    if not feedback_service_base_url:
        logger.debug("未配置外部反馈 API URL，跳过从外部 API 获取反馈")
        return []

    external_api_url = f"{feedback_service_base_url}/feedback/trace"
    headers = {
        "Content-Type": "application/json",
        "Authorization": "Bearer readonly-admin-token"
    }
    timeout = _get_feedback_api_timeout()

    try:
        async with httpx.AsyncClient(timeout=timeout) as client:
            response = await client.get(
                external_api_url,
                params={"trace_id": trace_id},
                headers=headers
            )
            response.raise_for_status()
            data = response.json()

            # 假设响应格式为 {"feedbacks": [...]} 或直接是列表
            if isinstance(data, dict) and "feedbacks" in data:
                feedbacks = data["feedbacks"]
            elif isinstance(data, list):
                feedbacks = data
            else:
                logger.warning(f"外部 API 返回格式未知: {data}")
                return []

            logger.info(f"从外部 API 获取反馈成功: trace_id={trace_id}, 条数={len(feedbacks)}")
            return feedbacks

    except httpx.TimeoutException:
        logger.warning(f"外部 API 请求超时: {external_api_url}")
        return []
    except httpx.ConnectError as e:
        logger.warning(f"外部 API 连接失败: {external_api_url} - {e}")
        return []
    except httpx.HTTPStatusError as e:
        logger.error(f"外部 API 请求失败: status={e.response.status_code}, detail={e.response.text}")
        return []
    except Exception as e:
        logger.error(f"外部 API 请求异常: {type(e).__name__}: {e}")
        return []

async def fetch_feedbacks_by_workflow_from_external_api(workflow_name: str) -> list[dict]:
    """
    从外部反馈服务 API 获取反馈数据 (按 workflow_name)

    Args:
        workflow_name: 工作流名称

    Returns:
        反馈数据列表
    """
    feedback_service_base_url = _get_feedback_api_base()
    if not feedback_service_base_url:
        logger.debug("未配置外部反馈 API URL，跳过从外部 API 获取反馈")
        return []

    external_api_url = f"{feedback_service_base_url}/feedback/workflow"
    headers = {
        "Content-Type": "application/json",
        "Authorization": "Bearer readonly-admin-token"
    }
    timeout = _get_feedback_api_timeout()

    try:
        async with httpx.AsyncClient(timeout=timeout) as client:
            response = await client.get(
                external_api_url,
                params={"workflow_name": workflow_name},
                headers=headers
            )
            response.raise_for_status()
            data = response.json()

            if isinstance(data, dict) and "feedbacks" in data:
                feedbacks = data["feedbacks"]
            elif isinstance(data, list):
                feedbacks = data
            else:
                logger.warning(f"外部 API 返回格式未知: {data}")
                return []

            logger.info(f"从外部 API 获取反馈成功: workflow_name={workflow_name}, 条数={len(feedbacks)}")
            return feedbacks

    except httpx.TimeoutException:
        logger.warning(f"外部 API 请求超时: {external_api_url}")
        return []
    except httpx.ConnectError as e:
        logger.warning(f"外部 API 连接失败: {external_api_url} - {e}")
        return []
    except httpx.HTTPStatusError as e:
        logger.error(f"外部 API 请求失败: status={e.response.status_code}, detail={e.response.text}")
        return []
    except Exception as e:
        logger.error(f"外部 API 请求异常: {type(e).__name__}: {e}")
        return []

async def fetch_all_feedbacks_from_external_api() -> list[dict]:
    """
    从外部反馈服务 API 获取所有反馈数据

    Returns:
        反馈数据列表
    """
    feedback_service_base_url = _get_feedback_api_base()
    if not feedback_service_base_url:
        logger.debug("未配置外部反馈 API URL，跳过从外部 API 获取反馈")
        return []

    external_api_url = f"{feedback_service_base_url}/feedback/all"
    headers = {
        "Content-Type": "application/json",
        "Authorization": "Bearer readonly-admin-token"
    }
    timeout = _get_feedback_api_timeout()

    try:
        async with httpx.AsyncClient(timeout=timeout) as client:
            response = await client.get(
                external_api_url,
                headers=headers
            )
            response.raise_for_status()
            data = response.json()

            if isinstance(data, dict) and "feedbacks" in data:
                feedbacks = data["feedbacks"]
            elif isinstance(data, list):
                feedbacks = data
            else:
                logger.warning(f"外部 API 返回格式未知: {data}")
                return []

            logger.info(f"从外部 API 获取所有反馈成功: 条数={len(feedbacks)}")
            return feedbacks

    except httpx.TimeoutException:
        logger.warning(f"外部 API 请求超时: {external_api_url}")
        return []
    except httpx.ConnectError as e:
        logger.warning(f"外部 API 连接失败: {external_api_url} - {e}")
        return []
    except httpx.HTTPStatusError as e:
        logger.error(f"外部 API 请求失败: status={e.response.status_code}, detail={e.response.text}")
        return []
    except Exception as e:
        logger.error(f"外部 API 请求异常: {type(e).__name__}: {e}")
        return []

def merge_feedbacks(local_feedbacks: list[dict], external_feedbacks: list[dict]) -> list[dict]:
    """
    合并本地反馈和外部 API 反馈，去除重复

    Args:
        local_feedbacks: 本地数据库反馈列表
        external_feedbacks: 外部 API 反馈列表

    Returns:
        合并后的反馈列表（去重，按创建时间倒序）
    """
    # 使用 feedback_id 作为去重键
    seen_ids = set()
    merged = []

    # 先添加本地反馈
    for feedback in local_feedbacks:
        feedback_id = feedback.get("feedback_id")
        if feedback_id and feedback_id not in seen_ids:
            seen_ids.add(feedback_id)
            merged.append(feedback)

    # 再添加外部反馈（避免重复）
    for feedback in external_feedbacks:
        feedback_id = feedback.get("feedback_id")
        if feedback_id and feedback_id not in seen_ids:
            seen_ids.add(feedback_id)
            merged.append(feedback)

    # 按 created_at 倒序排列
    def sort_key(x):
        created_at = x.get("created_at")
        # 如果是 datetime 对象，直接返回
        if isinstance(created_at, datetime):
            return created_at
        # 如果是字符串，尝试转换为 datetime
        if isinstance(created_at, str):
            try:
                return datetime.fromisoformat(created_at.replace('Z', '+00:00'))
            except (ValueError, AttributeError):
                return datetime.min
        # 其他情况返回最小时间
        return datetime.min

    try:
        merged.sort(key=sort_key, reverse=True)
    except Exception as e:
        logger.warning(f"排序反馈失败: {e}")

    return merged

@router.post("/", response_model=FeedbackResponse, summary="创建工作流反馈")
async def create_feedback(feedback: FeedbackCreate, background_tasks: BackgroundTasks, request: Request):
    """
    创建工作流执行完成后的用户反馈
    - **trace_id**: trace_id
    - **service_name**: 服务名称（可选，如果未提供会从 trace API 获取）
    - **workflow_name**: 工作流名称（可选，如果未提供会从 trace API 获取）
    - **rating**: 可选的评分 (1-5)
    - **comment**: 可选的用户评论
    - **metadata**: 可选的额外元数据
    """
    try:
        service_name = feedback.service_name
        workflow_name = feedback.workflow_name

        # 如果 service_name 或 workflow_name 缺失，从 trace API 获取
        if not service_name or not workflow_name:
            logger.info(f"service_name 或 workflow_name 缺失，尝试从 trace API 获取: trace_id={feedback.trace_id}")
            trace_info = await fetch_trace_info_from_trace(feedback.trace_id)

            if not service_name and trace_info.get("service_name"):
                service_name = trace_info.get("service_name")
            if not workflow_name and trace_info.get("workflow_name"):
                workflow_name = trace_info.get("workflow_name")

        # 使用空字符串作为默认值
        if not service_name:
            logger.warning(f"无法获取 service_name: trace_id={feedback.trace_id}，使用空字符串")
            service_name = ""
        if not workflow_name:
            logger.warning(f"无法获取 workflow_name: trace_id={feedback.trace_id}，使用空字符串")
            workflow_name = ""

        # 保存到数据库
        feedback_data = await feedback_storage.create_feedback(
            trace_id=feedback.trace_id,
            service_name=service_name,
            workflow_name=workflow_name,
            rating=feedback.rating,
            comment=feedback.comment,
            metadata=feedback.metadata,
        )

        # 从 request.state 获取用户认证信息
        user_id = getattr(request.state, "user_id", None)
        auth_token = getattr(request.state, "auth_token", None)

        # 检查是否已经被转发过（从请求头中获取转发标记）
        is_forwarded = request.headers.get("X-Feedback-Forwarded") == "true"

        # 后台任务转发到外部 API (不阻塞响应)
        background_tasks.add_task(forward_feedback_to_api, feedback_data, user_id, auth_token, is_forwarded)

        return FeedbackResponse(**feedback_data)
    except Exception as e:
        logger.error(f"创建反馈失败: {e}")
        raise HTTPException(status_code=500, detail=f"创建反馈失败: {str(e)}")

@router.get("/{feedback_id}", response_model=FeedbackResponse, summary="获取单个反馈")
async def get_feedback(feedback_id: str):
    """
    根据反馈ID获取反馈详情

    - **feedback_id**: 反馈ID
    """
    feedback = await feedback_storage.get_feedback(feedback_id)
    if not feedback:
        raise HTTPException(status_code=404, detail="反馈不存在")
    return FeedbackResponse(**feedback)

@router.get("/", response_model=FeedbackListResponse, summary="获取反馈列表")
async def list_feedbacks(
    page: int = Query(1, ge=1, description="页码（从1开始）"),
    page_size: int = Query(20, ge=1, le=100, description="每页条目数（1-100）"),
    trace_id: Optional[str] = Query(None, description="按trace_id筛选"),
    workflow_name: Optional[str] = Query(None, description="按工作流名称筛选"),
):
    """
    获取反馈列表,支持按trace_id或工作流名称筛选，同时会从外部 API 合并数据

    - **page**: 页码，从1开始（默认1）
    - **page_size**: 每页条目数，1-100（默认20）
    - **trace_id**: 可选,按trace_id筛选（会同时查询外部 API）
    - **workflow_name**: 可选,按工作流名称筛选
    """
    try:
        # 获取本地反馈
        if trace_id:
            local_feedbacks = await feedback_storage.get_feedbacks_by_trace(trace_id)
            # 如果指定了 trace_id，同时从外部 API 获取反馈
            external_feedbacks = await fetch_feedbacks_from_external_api(trace_id)
            # 合并本地和外部反馈
            feedbacks = merge_feedbacks(local_feedbacks, external_feedbacks)
        elif workflow_name:
            local_feedbacks = await feedback_storage.get_feedbacks_by_workflow(workflow_name)
            # 如果指定了 workflow_name，同时从外部 API 获取反馈
            external_feedbacks = await fetch_feedbacks_by_workflow_from_external_api(workflow_name)
            # 合并本地和外部反馈
            feedbacks = merge_feedbacks(local_feedbacks, external_feedbacks)
        else:
            local_feedbacks = await feedback_storage.get_all_feedbacks()
            # 获取所有反馈时，同时从外部 API 获取反馈
            external_feedbacks = await fetch_all_feedbacks_from_external_api()
            # 合并本地和外部反馈
            feedbacks = merge_feedbacks(local_feedbacks, external_feedbacks)

        # 计算分页
        total = len(feedbacks)
        start_idx = (page - 1) * page_size
        end_idx = start_idx + page_size
        paginated_feedbacks = feedbacks[start_idx:end_idx]

        return FeedbackListResponse(
            total=total,
            feedbacks=[FeedbackResponse(**f) for f in paginated_feedbacks]
        )
    except Exception as e:
        logger.error(f"获取反馈列表失败: {e}")
        raise HTTPException(status_code=500, detail=f"获取反馈列表失败: {str(e)}")

@router.delete("/{feedback_id}", summary="删除反馈")
async def delete_feedback(feedback_id: str):
    """
    删除指定的反馈

    - **feedback_id**: 反馈ID
    """
    success = await feedback_storage.delete_feedback(feedback_id)
    if not success:
        raise HTTPException(status_code=404, detail="反馈不存在")
    return {"message": "反馈删除成功", "feedback_id": feedback_id}



from __future__ import annotations
import traceback
from loguru import logger

from service_forge.workflow.node import Node
from service_forge.workflow.port import Port
from service_forge.workflow.registry.prompt_registry import load_prompt
from service_forge.llm import Model
from service_forge.llm.cleaner.json_cleaner import JsonCleaner

class LLMNode(Node):
    DEFAULT_INPUT_PORTS = [
        Port("prompt", str),
        Port("system_prompt", str),
        Port("model", Model),
        Port("temperature", float),
        Port("cleaner", str),
    ]

    DEFAULT_OUTPUT_PORTS = [
        Port("content", str),
        Port("error", str),
    ]

    def __init__(
        self, 
        name: str,
    ) -> None:
        super().__init__(
            name,
        )

    async def _run(
        self,
        prompt: str,
        system_prompt: str,
        model: Model,
        temperature: float,
        cleaner: str,
    ) -> None:
        try:
            system_prompt = load_prompt(system_prompt)
            prompt = load_prompt(prompt)

            response = await self.llm.acompletion(
                input=prompt,
                system_prompt=system_prompt,
                model=model,
                temperature=temperature,
            )

            content = response.choices[0].message.content

            if cleaner:
                for name in cleaner.split('|'):
                    if name.strip().lstrip() == 'json':
                        content = JsonCleaner().clean(content)
        
            self.activate_output_edges('content', content)
        except Exception as e:
            logger.error(f"Error when running LLMNode {self.name}: {e}, traceback: {traceback.format_exc()}")
            self.activate_output_edges('error', str(e))


"""change_user_id_to_string

Revision ID: 3c2f486fa1d4
Revises: 78e0ab93b1e5
Create Date: 2026-01-09 23:14:18.264394

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '3c2f486fa1d4'
down_revision: Union[str, Sequence[str], None] = '78e0ab93b1e5'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    # Change column type from Integer to String
    # postgresql_using will automatically convert integer values to text
    op.alter_column('tag', 'user_id',
                    type_=sa.String(255),
                    existing_type=sa.Integer(),
                    existing_nullable=True,
                    postgresql_using='user_id::text')


def downgrade() -> None:
    """Downgrade schema."""
    # Change column type from String back to Integer
    # postgresql_using will automatically convert text values to integer
    # Note: This will fail if any user_id values are not valid integers
    op.alter_column('tag', 'user_id',
                    type_=sa.Integer(),
                    existing_type=sa.String(255),
                    existing_nullable=True,
                    postgresql_using='user_id::integer')



import os
import asyncio
from logging.config import fileConfig

from sqlalchemy import pool
from sqlalchemy.ext.asyncio import async_engine_from_config

from alembic import context

from dotenv import load_dotenv

from service_forge.sft.config.sf_metadata import load_metadata
from service_forge.db.database import DatabaseManager

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

metadata = load_metadata("./sf-meta.yaml")

database = DatabaseManager.from_config(config_path=metadata.service_config)
url = database.get_default_postgres_database().database_url

config.set_main_option("sqlalchemy.url", url)

# Interpret the config file for Python logging.
# This line sets up loggers basically.
if config.config_file_name is not None:
    fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# Import all models to ensure they are registered with Base.metadata
import sys
import importlib
import inspect
from pathlib import Path

# Add the project root to the path so we can import models
sys.path.insert(0, str(Path(__file__).resolve().parents[2]))

# Import Base first
from model.db import Base

def import_all_base_models():
    db_path = Path(__file__).resolve().parents[1] / "model" / "db"
    
    for file_path in sorted(db_path.glob("*.py")):
        if file_path.name == "__init__.py":
            continue
        
        module_name = file_path.stem
        try:
            print(f"Importing module: model.db.{module_name}")
            module = importlib.import_module(f"model.db.{module_name}")
            
            for name, obj in inspect.getmembers(module, inspect.isclass):
                if (issubclass(obj, Base) and 
                    obj is not Base and
                    obj.__module__ == module.__name__):
                    if hasattr(obj, '__table__'):
                        _ = obj.__table__
        except Exception as e:
            import warnings
            warnings.warn(f"Failed to import model.db.{module_name}: {e}")

import_all_base_models()

# Set target_metadata for autogenerate support
target_metadata = Base.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def do_run_migrations(connection):
    context.configure(connection=connection, target_metadata=target_metadata)

    with context.begin_transaction():
        context.run_migrations()


async def run_async_migrations():
    """In this scenario we need to create an Engine
    and associate a connection with the context.

    """

    connectable = async_engine_from_config(
        config.get_section(config.config_ini_section),
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    async with connectable.connect() as connection:
        await connection.run_sync(do_run_migrations)

    await connectable.dispose()


def run_migrations_online():
    """Run migrations in 'online' mode."""
    
    # Check if we have a connection provided via the attributes (from programmatic use)
    connectable = config.attributes.get("connection", None)

    if connectable is None:
        # No connection provided, run async migrations
        asyncio.run(run_async_migrations())
    else:
        # Connection provided programmatically, use it directly
        do_run_migrations(connectable)


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()


import json
import os
import uuid
import tempfile
from fastapi import APIRouter, HTTPException, UploadFile, File, Form
from fastapi.params import Body, Path
from fastapi.responses import JSONResponse
from loguru import logger
from typing import Optional, TYPE_CHECKING, Dict, Any
from pydantic import BaseModel
from omegaconf import OmegaConf
from service_forge.current_service import get_service

service_router = APIRouter(prefix="/sdk/service", tags=["service"])

class WorkflowStatusResponse(BaseModel):
    name: str
    version: str
    description: str
    workflows: list[dict]

class WorkflowActionResponse(BaseModel):
    workflow_id: Optional[str] = None
    success: bool
    message: str
    task_id: Optional[str] = None

@service_router.get("/status", response_model=WorkflowStatusResponse)
async def get_service_status():
    service = get_service()
    if service is None:
        raise HTTPException(status_code=503, detail="Service not initialized")
    # 排除调试版本
    try:
        status = service.get_service_status(exclude_debug=True)
        return status
    except Exception as e:
        logger.error(f"Error getting service status: {e}")
        raise HTTPException(status_code=500, detail=str(e))

@service_router.get("/workflow/{workflow_id}/status", response_model=dict)
def get_workflow_data(workflow_id: str):
    service = get_service()
    if service is None:
        raise HTTPException(status_code=503, detail="Service not initialized")

    try:
        status = service.get_workflow_status(workflow_id)
        return status
    except Exception as e:
        logger.error(f"Error getting service status: {e}")
        raise HTTPException(status_code=500, detail=str(e))

@service_router.post("/workflow/{workflow_id}/start", response_model=WorkflowActionResponse)
async def start_workflow(workflow_id: str):
    service = get_service()
    if service is None:
        raise HTTPException(status_code=503, detail="Service not initialized")
    
    try:
        success = service.start_workflow_by_id(workflow_id)
        if success:
            return WorkflowActionResponse(success=True, message=f"Workflow {workflow_id} started successfully")
        else:
            return WorkflowActionResponse(success=False, message=f"Failed to start workflow {workflow_id}")
    except Exception as e:
        logger.error(f"Error starting workflow {workflow_id}: {e}")
        raise HTTPException(status_code=500, detail=str(e))

@service_router.post("/workflow/{workflow_id}/stop", response_model=WorkflowActionResponse)
async def stop_workflow(workflow_id: str):
    service = get_service()
    if service is None:
        raise HTTPException(status_code=503, detail="Service not initialized")
    
    try:
        success = await service.stop_workflow_by_id(workflow_id)
        if success:
            return WorkflowActionResponse(success=True, message=f"Workflow {workflow_id} stopped successfully")
        else:
            return WorkflowActionResponse(success=False, message=f"Failed to stop workflow {workflow_id}")
    except Exception as e:
        logger.error(f"Error stopping workflow {workflow_id}: {e}")
        raise HTTPException(status_code=500, detail=str(e))

class TriggerWorkflowRequest(BaseModel):
    kwargs: Optional[Dict[str, Any]] = {}

@service_router.post("/workflow/{workflow_id}/trigger", response_model=WorkflowActionResponse)
async def trigger_workflow(workflow_id: str = Path(...), request_body: TriggerWorkflowRequest = Body(...)):
    service = get_service()
    if service is None:
        raise HTTPException(status_code=503, detail="Service not initialized")

    try:
        task_id = service.trigger_workflow_by_id(workflow_id, "", None, **request_body.kwargs)   # Trigger Name实际上没有使用
        if task_id is not None:
            return WorkflowActionResponse(workflow_id=workflow_id, task_id=str(task_id), success=True, message=f"Workflow {workflow_id} triggered successfully with task_id {task_id}")
        else:
            return WorkflowActionResponse(workflow_id=workflow_id, success=False, message=f"Failed to trigger workflow {workflow_id}")
    except Exception as e:
        logger.error(f"Error triggering workflow {workflow_id}: {e}")
        raise HTTPException(status_code=500, detail=str(e))


@service_router.post("/workflow/upload", response_model=WorkflowActionResponse)
async def upload_workflow_config(
    file: Optional[UploadFile] = File(None),
    config_content: Optional[str] = Form(None),
):
    service = get_service()
    if service is None:
        raise HTTPException(status_code=503, detail="Service not initialized")
    
    if file is None and config_content is None:
        raise HTTPException(status_code=400, detail="Either file or config_content must be provided")
    
    if file is not None and config_content is not None:
        raise HTTPException(status_code=400, detail="Cannot provide both file and config_content")
    
    temp_file_path = None
    try:
        if file is not None:
            if not file.filename or not file.filename.endswith(('.yaml', '.yml')):
                raise HTTPException(status_code=400, detail="Only YAML files are supported")
            
            suffix = '.yaml' if file.filename.endswith('.yaml') else '.yml'
            with tempfile.NamedTemporaryFile(mode='wb', delete=False, suffix=suffix) as temp_file:
                content = await file.read()
                temp_file.write(content)
                temp_file_path = temp_file.name
           
            workflow_id = await service.load_workflow_from_config(config_path=temp_file_path)
        else:
            try:
                config = OmegaConf.to_object(OmegaConf.create(config_content))
            except Exception as e:
                raise HTTPException(status_code=400, detail=f"Invalid YAML format: {str(e)}")
            
            workflow_id = await service.load_workflow_from_config(config=config)
        
        if workflow_id:
            return WorkflowActionResponse(
                workflow_id=str(workflow_id),
                success=True,
                message=f"Workflow configuration uploaded and loaded successfully"
            )
        else:
            raise HTTPException(status_code=500, detail="Failed to load workflow configuration")
    
    except HTTPException:
        raise
    except Exception as e:
        logger.error(f"Error uploading workflow config: {e}")
        raise HTTPException(status_code=500, detail=str(e))
    
    finally:
        if temp_file_path and os.path.exists(temp_file_path):
            try:
                os.unlink(temp_file_path)
            except Exception as e:
                logger.warning(f"Failed to delete temp file {temp_file_path}: {e}")




from ..llm_cleaner import LLMCleaner
from typing import Any
import json
import re

class JsonCleaner(LLMCleaner):
    def __init__(self):
        pass

    def clean(self, content: str) -> Any:
        pattern = r'```json\s*(.*?)\s*```'
        match = re.search(pattern, content, re.DOTALL)
        
        if match:
            json_str = match.group(1).strip()
            return json.loads(json_str)
        else:
            raise ValueError("No JSON code block found in content")



from __future__ import annotations
from pydantic import BaseModel

from service_forge.utils.register import Register

class SfBaseModel(BaseModel):
    def __init_subclass__(cls) -> None:
        if cls.__name__ not in ["SfBaseModel"]:
            sf_basemodel_register.register(cls.__name__, cls)
        return super().__init_subclass__()

sf_basemodel_register = Register[SfBaseModel]()


from __future__ import annotations

import json
from contextlib import asynccontextmanager
from opentelemetry.trace import SpanKind, Tracer
from opentelemetry.context import Context
from opentelemetry.trace import Span
from service_forge.workflow.trigger_event import TriggerEvent
from .execution_context import (
    ExecutionContext,
    get_current_context,
    reset_current_context,
    set_current_context,
)

from typing import TYPE_CHECKING, Any

from pydantic import BaseModel
if TYPE_CHECKING:
    from service_forge.workflow.workflow import Workflow
    from service_forge.workflow.node import Node

@asynccontextmanager
async def workflow_span(
    tracer: Tracer,
    *,
    workflow: Workflow,
    parent_context: Context,
    span_kind: SpanKind = SpanKind.INTERNAL,
    extra_attributes: dict | None = None,
):
    with tracer.start_as_current_span(
        name=f"Workflow {workflow.name}",
        context=parent_context,
        kind=span_kind,
    ) as span:
        span.set_attribute("workflow.name", workflow.name)
        span.set_attribute("workflow.task_id", str(workflow.trigger_event.task_id))
        if extra_attributes:
            for k, v in extra_attributes.items():
                span.set_attribute(k, v)
        yield span

def _serialize_value(value: Any) -> str:
    if value is None:
        return "null"
    if isinstance(value, (str, int, float, bool)):
        return str(value)
    
    if isinstance(value, BaseModel):
        if hasattr(value, 'model_dump'):
            try:
                return json.dumps(value.model_dump(), default=str, ensure_ascii=False)
            except (TypeError, ValueError):
                pass
        elif hasattr(value, 'dict'):
            try:
                return json.dumps(value.dict(), default=str, ensure_ascii=False)
            except (TypeError, ValueError):
                pass
    
    try:
        return json.dumps(value, default=str, ensure_ascii=False)
    except (TypeError, ValueError):
        return str(value)

@asynccontextmanager
async def node_span(
    tracer: Tracer,
    *,
    node: Node,
    span_kind: SpanKind = SpanKind.INTERNAL,
    extra_attributes: dict | None = None,
):
    with tracer.start_as_current_span(
        name=f"Node {node.name}",
        kind=span_kind,
    ) as span:
        span.set_attribute("node.name", node.name)
        node_input_attributes(span, node)

        if extra_attributes:
            for k, v in extra_attributes.items():
                span.set_attribute(k, v)
        
        try:
            yield span
        finally:
            node_output_attributes(span, node)

def node_input_attributes(span: Span, node: Node):
    for input_port in node.input_ports:
        value = None
        if input_port in node.input_variables:
            value = node.input_variables[input_port]
        elif input_port.is_prepared and input_port.value is not None:
            value = input_port.value
        
        if value is not None:
            attr_name = f"node.input.{input_port.name}"
            span.set_attribute(attr_name, _serialize_value(value))

def node_output_attributes(span: Span, node: Node):
    for output_port in node.output_ports:
        if output_port.value is not None:
            attr_name = f"node.output.{output_port.name}"
            span.set_attribute(attr_name, _serialize_value(output_port.value))


"""add_trace_id_and_span_id

Revision ID: 2312726d56e1
Revises: 3c2f486fa1d4
Create Date: 2026-01-16 23:38:45.417339

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '2312726d56e1'
down_revision: Union[str, Sequence[str], None] = '3c2f486fa1d4'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('tag', sa.Column('trace_id', sa.String(length=32), nullable=True, comment='OpenTelemetry trace_id'))
    op.add_column('tag', sa.Column('span_id', sa.String(length=16), nullable=True, comment='OpenTelemetry span_id'))
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('tag', 'span_id')
    op.drop_column('tag', 'trace_id')
    # ### end Alembic commands ###



from __future__ import annotations

import json
from contextlib import contextmanager
from contextvars import ContextVar, Token
from dataclasses import dataclass, field, replace
from typing import Any, Iterator, Mapping

from loguru import logger
from opentelemetry.context import Context as OtelContext
from opentelemetry.trace import Span

"""
ExecutionContext helper for tracing/logging/state propagation.

Usage constraints (async-aware):
- Entrypoints must call set_current_context/use_execution_context so spawned coroutines inherit the context; create tasks after setting it.
- ContextVar does not flow into run_in_executor/thread pools; set_current_context manually in threads if needed.
- Use with_state to copy + mutate state; instances are immutable to avoid concurrent writes.
- Logger/span are shared references; set a new ExecutionContext when switching spans to prevent cross-task leakage.
- State payload is masked for sensitive keys and capped at ~4KB after JSON serialization; oversize payload raises ValueError.
"""

STATE_SIZE_LIMIT_BYTES = 4096
SENSITIVE_KEYS = (
    "password",
    "passwd",
    "secret",
    "token",
    "api_key",
    "apikey",
    "authorization",
    "auth",
    "credential",
    "cookie",
)


def _mask_sensitive_value(key: str, value: Any) -> Any:
    lower_key = key.lower()
    if any(sensitive in lower_key for sensitive in SENSITIVE_KEYS):
        return "***"
    return value


def _validate_state_payload(state: Mapping[str, Any]) -> None:
    try:
        payload = json.dumps(state, default=str)
    except Exception:
        payload = str(state)
    if len(payload.encode("utf-8")) > STATE_SIZE_LIMIT_BYTES:
        raise ValueError(
            f"ExecutionContext state exceeds {STATE_SIZE_LIMIT_BYTES} bytes after serialization; "
            "store a compact summary instead of raw payloads."
        )


@dataclass(frozen=True)
class ExecutionContext:
    """Shared execution-scoped data passed across API, workflow, and node layers."""

    trace_context: OtelContext | None = None
    span: Span | None = None
    logger = logger
    state: Mapping[str, Any] = field(default_factory=dict)
    metadata: Mapping[str, Any] = field(default_factory=dict)

    def with_state(self, **updates: Any) -> ExecutionContext:
        merged_state = {
            key: _mask_sensitive_value(key, value)
            for key, value in {**self.state, **updates}.items()
        }
        _validate_state_payload(merged_state)
        return replace(self, state=merged_state)


current_context: ContextVar[ExecutionContext | None] = ContextVar(
    "current_execution_context",
    default=None,
)


def get_current_context(
    default: ExecutionContext | None = None,
) -> ExecutionContext | None:
    context = current_context.get()
    if context is None:
        return default
    return context


def set_current_context(context: ExecutionContext | None) -> Token:
    return current_context.set(context)


def reset_current_context(token: Token) -> None:
    current_context.reset(token)


@contextmanager
def use_execution_context(context: ExecutionContext) -> Iterator[ExecutionContext]:
    token = set_current_context(context)
    try:
        yield context
    finally:
        reset_current_context(token)



from typing import Any
from service_forge.workflow.node import Node
from service_forge.workflow.port import Port

class CreateObjectNode(Node):
    DEFAULT_INPUT_PORTS = [
        Port("object_type", type),
        Port("fields", tuple[str, Any], is_extended=True),
    ]

    DEFAULT_OUTPUT_PORTS = [
        Port("object", Any),
    ]

    def __init__(self, name: str) -> None:
        super().__init__(name)

    def _run(
        self,
        object_type: type,
        fields: list[tuple[int, tuple[str, Any]]],
    ) -> Any:
        if object_type == dict:
            for _, f in fields:
                object[f[0]] = f[1]
        elif object_type == list:
            for _, f in fields:
                object.append(f[1])
        else:
            fields = {f[0]: f[1] for _, f in fields}
            object = object_type(**fields)

        self.activate_output_edges('object', object)



from service_forge.workflow.registry.sf_base_model import SfBaseModel

class QueryTagsModel(SfBaseModel):
    ids: str = ""
    page: int = 1
    page_size: int = 10
    sort_by: str = "created_at"
    order: str = "desc"


#!/usr/bin/env python3
from __future__ import annotations

import json
import sys
from typing import Any
from urllib.parse import urlparse
from urllib.request import urlopen
from datetime import datetime


def load_openapi(source: str) -> dict:
    """Load openapi.json from file path or URL"""
    parsed = urlparse(source)
    if parsed.scheme in ("http", "https"):
        with urlopen(source) as resp:
            return json.load(resp)
    else:
        with open(source, "r", encoding="utf-8") as f:
            return json.load(f)


def py_type(schema: dict) -> str:
    """OpenAPI schema -> Python type string"""
    if not schema:
        return "Any"
    
    if not isinstance(schema, dict):
        return "Any"

    if "$ref" in schema:
        return schema["$ref"].split("/")[-1]

    if "anyOf" in schema:
        types = []
        nullable = False
        for s in schema["anyOf"]:
            if s.get("type") == "null":
                nullable = True
            else:
                types.append(py_type(s))
        t = " | ".join(sorted(set(types))) or "Any"
        return f"{t} | None" if nullable else t

    t = schema.get("type")

    if t == "string":
        if schema.get("format") == "date-time":
            return "datetime"
        return "str"

    if t == "integer":
        return "int"

    if t == "number":
        return "float"

    if t == "boolean":
        return "bool"

    if t == "array":
        return f"list[{py_type(schema.get('items', {}))}]"

    if t == "object":
        if "additionalProperties" in schema:
            additional_props = schema["additionalProperties"]
            if isinstance(additional_props, bool):
                return "dict[str, Any]"
            return f"dict[str, {py_type(additional_props)}]"
        return "dict[str, Any]"

    return "Any"


def gen_model(name: str, schema: dict) -> str:
    required = set(schema.get("required", []))
    props = schema.get("properties", {})

    lines = [f"class {name}(BaseModel):"]

    if not props:
        lines.append("    pass")
        return "\n".join(lines)

    for prop, pschema in props.items():
        t = py_type(pschema)
        default = "" if prop in required else " = None"
        lines.append(f"    {prop}: {t}{default}")

    return "\n".join(lines)


def main():
    if len(sys.argv) != 2:
        print("Usage: python gen_models.py <openapi.json | url>", file=sys.stderr)
        sys.exit(1)

    spec = load_openapi(sys.argv[1])
    schemas = spec.get("components", {}).get("schemas", {})

    print("from __future__ import annotations")
    print("from pydantic import BaseModel")
    print("from typing import Any")
    print("from datetime import datetime")
    print()

    for name, schema in schemas.items():
        print(gen_model(name, schema))
        print()


if __name__ == "__main__":
    main()


from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Optional,
    Union,
    cast,
)

from loguru import logger
from opentelemetry import trace

from .edge import Edge
from .port import Port
from .context import Context
from .workflow_callback import CallbackEvent
from .workflow_clone import node_clone
from ..db.database import DatabaseManager, PostgresDatabase, MongoDatabase, RedisDatabase
from ..llm import SfLLM
from ..utils.register import Register
from ..sft.util.name_util import get_url, get_service_name
from ..sft.config.sf_metadata import metadata

if TYPE_CHECKING:
    from .workflow import Workflow

class Node(ABC):
    DEFAULT_INPUT_PORTS: list[Port] = []
    DEFAULT_OUTPUT_PORTS: list[Port] = []

    CLASS_NOT_REQUIRED_TO_REGISTER = ["Node"]
    AUTO_FILL_INPUT_PORTS = []

    def __init__(
        self,
        name: str,
        context: Optional[Context] = None,
        input_edges: Optional[list[Edge]] = None,
        output_edges: Optional[list[Edge]] = None,
        input_ports: list[Port] = DEFAULT_INPUT_PORTS,
        output_ports: list[Port] = DEFAULT_OUTPUT_PORTS,
        query_user: Optional[Callable[[str, str], Awaitable[str]]] = None,
    ) -> None:
        from .workflow_group import WorkflowGroup

        self.name = name
        self.input_edges = [] if input_edges is None else input_edges
        self.output_edges = [] if output_edges is None else output_edges
        self.input_ports = input_ports
        self.output_ports = output_ports
        self.workflow: Optional[Workflow] = None
        self.query_user = query_user
        self.sub_workflows: Optional[WorkflowGroup] = None

        # runtime variables
        self.context = context
        self.input_variables: dict[Port, Any] = {}
        self.num_activated_input_edges = 0
        self._tracer = trace.get_tracer("service_forge.node")

    @property
    def default_postgres_database(self) -> PostgresDatabase | None:
        return self.database_manager.get_default_postgres_database()

    @property
    def default_mongo_database(self) -> MongoDatabase | None:
        return self.database_manager.get_default_mongo_database()

    @property
    def default_redis_database(self) -> RedisDatabase | None:
        return self.database_manager.get_default_redis_database()
    
    @property
    def database_manager(self) -> DatabaseManager:
        if self.workflow is None:
            raise ValueError("Workflow is not set yet.")
        return self.workflow.database_manager
    
    @property
    def llm(self) -> SfLLM:
        if self.workflow is None:
            raise ValueError("Workflow is not set yet.")
        return self.workflow.llm

    @property
    def global_context(self) -> Context:
        return self.workflow.global_context

    def get_url(
        self,
        service_name: str = "",
        version: str = "",
        path: str = "",
        subdomain: str = "",
        path_params: dict = None,
        internal: bool = None,
    ) -> str:
        if internal is None:
            internal = metadata.mode != "debug"
        if path_params:
            path = path.format(**path_params)
        if internal:
            if service_name == 'entry':
                return f"http://entry-service:80/{path.lstrip('/')}"
            elif service_name == '' and subdomain == 'file':
                return f"http://seaweedfs-volume:8080/{path.lstrip('/')}"
            return f"http://{get_service_name(service_name, version)}:80/{path.lstrip('/')}"
        if self.workflow is None or self.workflow.entry_config is None or self.workflow.entry_config.url is None:
            logger.error("Entry config is not set, please set the entry config in the service.yaml file.")
            return ""
        return get_url(entry_url=self.workflow.entry_config.url, service_name=service_name, version=version, path=path, subdomain=subdomain)

    def __init_subclass__(cls) -> None:
        if cls.__name__ not in Node.CLASS_NOT_REQUIRED_TO_REGISTER:
            # TODO: Register currently stores class objects; clarify Register typing vs instance usage.
            node_register.register(cls.__name__, cls)
        return super().__init_subclass__()

    def _query_user(self, prompt: str) -> Awaitable[str]:
        assert self.query_user
        return self.query_user(self.name, prompt)

    def variables_to_params(self) -> dict[str, Any]:
        params = {port.name: self.input_variables[port] for port in self.input_variables.keys() if not port.is_extended_generated}
        for port in self.input_variables.keys():
            if port.is_extended_generated:
                if port.get_extended_name() not in params:
                    params[port.get_extended_name()] = []
                params[port.get_extended_name()].append((port.get_extended_index(), self.input_variables[port]))
                params[port.get_extended_name()].sort()
        return params

    def get_trace_id(self) -> str | None:
        span = trace.get_current_span()
        if span and span.get_span_context().is_valid:
            return format(span.get_span_context().trace_id, '032x')
        return None

    def get_span_id(self) -> str | None:
        span = trace.get_current_span()
        if span and span.get_span_context().is_valid:
            return format(span.get_span_context().span_id, '016x')
        return None

    def is_trigger(self) -> bool:
        from .trigger import Trigger
        return isinstance(self, Trigger)

    # TODO: maybe add a function before the run function?

    @abstractmethod
    async def _run(self, **kwargs) -> Union[None, AsyncIterator]:
        ...

    async def clear(self) -> None:
        ...
    
    def run(self) -> Union[None, AsyncIterator]:
        for key in list(self.input_variables.keys()):
            if key and key.name[0].isupper():
                del self.input_variables[key]
        logger.debug(f"Running node: {self.name} with input variables: {self.variables_to_params()}")
        result = self._run(**self.variables_to_params())
        return result

    def get_input_port_by_name(self, name: str) -> Optional[Port]:
        for port in self.input_ports:
            if port.name == name:
                return port
        return None

    def get_output_port_by_name(self, name: str) -> Optional[Port]:
        for port in self.output_ports:
            if port.name == name:
                return port
        return None

    def try_create_extended_input_port(self, name: str) -> None:
        for port in self.input_ports:
            if port.is_extended and name.startswith(port.name + '_') and name[len(port.name + '_'):].isdigit():
                self.input_ports.append(Port(name=name, type=port.type, node=port.node, port=port.port, value=port.value, default=port.default, is_extended=False, is_extended_generated=True))

    def try_create_extended_output_port(self, name: str) -> None:
        for port in self.output_ports:
            if port.is_extended and name.startswith(port.name + '_') and name[len(port.name + '_'):].isdigit():
                self.output_ports.append(Port(name=name, type=port.type, node=port.node, port=port.port, value=port.value, default=port.default, is_extended=False, is_extended_generated=True))

    def num_input_ports(self) -> int:
        return sum(1 for port in self.input_ports if not port.is_extended)
    
    def is_ready(self) -> bool:
        return self.num_activated_input_edges == self.num_input_ports()

    def fill_input_by_name(self, port_name: str, value: Any) -> None:
        self.try_create_extended_input_port(port_name)
        port = self.get_input_port_by_name(port_name)
        if port is None:
            raise ValueError(f"{port_name} is not a valid input port.")
        self.fill_input(port, value)

    def fill_input(self, port: Port, value: Any) -> None:
        port.activate(value)

    def activate_output_edges(self, port: str | Port, data: Any) -> None:
        if isinstance(port, str):
            port = self.get_output_port_by_name(port)
        port.value = data
        for output_edge in self.output_edges:
            if output_edge.start_port == port:
                output_edge.end_port.activate(output_edge.format(data))

    # for trigger nodes
    def prepare_output_edges(self, port: Port, data: Any) -> None:
        if isinstance(port, str):
            port = self.get_output_port_by_name(port)
        for output_edge in self.output_edges:
            if output_edge.start_port == port:
                output_edge.end_port.prepare(output_edge.format(data))

    def trigger_output_edges(self, port: Port) -> None:
        if isinstance(port, str):
            port = self.get_output_port_by_name(port)
        for output_edge in self.output_edges:
            if output_edge.start_port == port:
                output_edge.end_port.trigger()

    # TODO: the result is outputed to the trigger now, maybe we should add a new function to output the result to the workflow
    def output_to_workflow(self, data: Any) -> None:
        if self.workflow and hasattr(self.workflow, "_handle_workflow_output"):
            handler = cast(
                Callable[[str, Any], None],
                getattr(self.workflow, "_handle_workflow_output"),
            )
            handler(self.name, data)
        else:
            logger.warning("Workflow output handler not set; skipping output dispatch.")

    def extended_output_name(self, name: str, index: int) -> str:
        return name + '_' + str(index)

    def _clone(self, context: Context) -> Node:
        return node_clone(self, context)

    async def stream_output(self, data: Any) -> None:
         await self.workflow.call_callbacks(CallbackEvent.ON_NODE_STREAM_OUTPUT, node=self, output=data)


node_register = Register[Node]()



from __future__ import annotations

from typing import TYPE_CHECKING, Any
from .workflow_clone import edge_clone

if TYPE_CHECKING:
    from .node import Node
    from .port import Port

class Edge:
    def __init__(
        self,
        start_node: Node,
        end_node: Node,
        start_port: Port,
        end_port: Port,
        formatter: callable | None = None,
    ) -> None:
        self.start_node = start_node
        self.end_node = end_node
        self.start_port = start_port
        self.end_port = end_port
        self.formatter = formatter

    def _clone(self, node_map: dict[Node, Node], port_map: dict[Port, Port]) -> Edge:
        return edge_clone(self, node_map, port_map)

    def format(self, data: Any) -> Any:
        if self.formatter is not None:
            return self.formatter(data)
        return data


from fastapi import FastAPI
import uvicorn
from fastapi import APIRouter
from loguru import logger
from typing import Optional
from urllib.parse import urlparse
from fastapi import HTTPException, Request, WebSocket, WebSocketException
from fastapi.middleware.cors import CORSMiddleware

from opentelemetry import context as otel_context_api
from opentelemetry import trace
from opentelemetry.propagate import extract
from opentelemetry.trace import SpanKind
from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor

from service_forge.sft.config.sf_metadata import load_metadata
from service_forge.sft.util.name_util import get_service_url_name
from service_forge.api.routers.meta_api.meta_api_router import meta_api_router
from service_forge.api.routers.trace.trace_router import trace_router
from service_forge.api.service_studio import studio_static_files
from service_forge.api.routers.websocket.websocket_router import websocket_router
from service_forge.api.routers.service.service_router import service_router
from service_forge.api.routers.feedback.feedback_router import router as feedback_router
from service_forge.sft.config.sf_metadata import metadata

def create_app(
    app: FastAPI | None = None,
    routers: list[APIRouter] | None = None,
    cors_origins: list[str] | None = None,
    root_path: str | None = None,
) -> FastAPI:
    if app is None:
        app = FastAPI(root_path=root_path)
        FastAPIInstrumentor.instrument_app(app)
    
    # Configure CORS middleware
    if cors_origins is None:
        cors_origins = ["*"]

    app.add_middleware(
        CORSMiddleware,
        allow_origins=cors_origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    # Include routers if provided
    if routers:
        for router in routers:
            app.include_router(router)
    
    # Always include WebSocket router
    app.include_router(websocket_router)

    # Include Feedback router
    app.include_router(feedback_router)
    
    # Always include Service router, Meta API Router, Trace Router, and Static Files
    app.include_router(service_router)
    app.include_router(meta_api_router)
    app.include_router(trace_router)

    app.mount('/sdk/studio', studio_static_files)

    @app.middleware("http")
    async def auth_middleware(request: Request, call_next):
        request.state.user_id = request.headers.get("X-User-ID") or "0"
        request.state.auth_token = request.headers.get("X-User-Token") or ""
        return await call_next(request)

    return app


async def start_fastapi_server(host: str, port: int):
    try:
        config = uvicorn.Config(
            fastapi_app,
            host=host,
            port=int(port),
            log_level="info",
            access_log=True
        )
        server = uvicorn.Server(config)
        await server.serve()
    except Exception as e:
        logger.error(f"Server error: {e}")
        raise

try:
    # metadata = load_metadata("sf-meta.yaml")
    if metadata.mode == "debug":
        fastapi_app = create_app(root_path=None)
    else:
        fastapi_app = create_app(root_path=f"/api/v1/{get_service_url_name(metadata.name, metadata.version)}")
except Exception as e:
    logger.warning(f"Failed to load metadata, using default configuration: {e}")
    fastapi_app = create_app(root_path=None)



import asyncio
from typing import AsyncIterator
from service_forge.workflow.node import Node
from service_forge.workflow.port import Port
from model.http.test_sse_model import TestSSEModel
import uuid
from typing import Any

class TestWebSocketNode(Node):
    DEFAULT_INPUT_PORTS = [
        Port("message", Any),
        Port("client_id", uuid.UUID),
    ]

    DEFAULT_OUTPUT_PORTS = [
        Port("result", str)
    ]

    def __init__(self, name: str):
        super().__init__(name)

    async def _run(self, message: Any, client_id: uuid.UUID) -> AsyncIterator[str]:
        print("client_id", client_id, 'data', message)
        self.activate_output_edges('result', "Done!")


from typing import Any
from loguru import logger
from sqlalchemy import select, inspect
from sqlalchemy.orm import InstrumentedAttribute
from service_forge.workflow.node import Node
from service_forge.workflow.port import Port

class DatabaseUpdateFieldsNode(Node):
    DEFAULT_INPUT_PORTS = [
        Port("model", Any),
        Port("conditions", str, is_extended=True),
        Port("fields", tuple[str, Any], is_extended=True),
    ]

    DEFAULT_OUTPUT_PORTS = [
        Port("data", Any),
    ]

    def __init__(self, name: str):
        super().__init__(name)

    async def _run(
        self, 
        model: Any,
        conditions: list[tuple[int, str]] = None,
        fields: list[tuple[int, tuple[str, Any]]] = None,
    ) -> None:
        if hasattr(model, '__table__') or hasattr(model, '__mapper__'):
            if isinstance(model, type):
                model_class = model
            else:
                model_class = type(model)
        else:
            raise ValueError(f"model must be a SQLAlchemy model instance or class, got {model}")
        
        condition_list = []
        if conditions:
            for _, cond_str in conditions:
                condition_list.append(cond_str)
        
        update_fields = {}
        if fields:
            # for _, field in fields:
            #     update_fields[field[0]] = field[1]
            for _, field in fields:
                if type(field) == str:
                    field = eval(field)
                update_fields[field[0]] = field[1]
        
        if not update_fields:
            raise ValueError("No fields provided for update")
        
        sqlalchemy_conditions = self._build_sqlalchemy_conditions(model_class, condition_list)
        
        session_factory = await self.default_postgres_database.get_session_factory()
        async with session_factory() as session:
            try:
                query = select(model_class)
                if sqlalchemy_conditions:
                    query = query.where(*sqlalchemy_conditions)
                
                result = await session.execute(query)
                matching_records = result.scalars().all()
                
                if not matching_records:
                    logger.warning(f"No records found matching conditions in {model_class.__tablename__}")
                    self.activate_output_edges(self.get_output_port_by_name('data'), [])
                    return
                
                for record in matching_records:
                    for field_name, field_value in update_fields.items():
                        if not hasattr(model_class, field_name):
                            logger.warning(f"Field '{field_name}' does not exist in model {model_class.__name__}, skipping")
                            continue
                        setattr(record, field_name, field_value)
                
                await session.commit()
                
                for record in matching_records:
                    await session.refresh(record)
                
                logger.info(f"✓ Updated {len(matching_records)} record(s) in {model_class.__tablename__}")
                
                self.activate_output_edges(self.get_output_port_by_name('data'), matching_records)
                    
            except Exception as e:
                logger.error(f"✗ Failed to update records: {e}")
                await session.rollback()
                raise
    
    def _build_sqlalchemy_conditions(self, model_class: type, condition_strings: list[str]) -> list:
        """Build SQLAlchemy conditions from condition strings
        
        Supports formats:
        - "Model.field == value" (full path)
        - "field == value" (simple equality)
        - "field != value" (inequality)
        - "field > value", "field < value", etc. (comparisons)
        """
        conditions = []
        
        for cond_str in condition_strings:
            if not cond_str or not isinstance(cond_str, str):
                continue
                
            try:
                cond_str = cond_str.strip()
                
                # Try full path format first: "Model.field == value"
                if '.' in cond_str and model_class.__name__ in cond_str:
                    try:
                        safe_dict = {
                            model_class.__name__: model_class,
                            '__builtins__': {},
                            'True': True,
                            'False': False,
                            'None': None,
                        }
                        condition = eval(cond_str, safe_dict)
                        if condition is not None:
                            conditions.append(condition)
                            continue
                    except Exception as e:
                        logger.debug(f"Could not evaluate full path condition '{cond_str}': {e}")
                
                # Parse simple conditions: "field operator value"
                operators = ['==', '!=', '>=', '<=', '>', '<', ' in ', ' not in ']
                for op in operators:
                    if op in cond_str:
                        parts = cond_str.split(op, 1)
                        if len(parts) == 2:
                            field_name = parts[0].strip()
                            value_str = parts[1].strip()
                            
                            if hasattr(model_class, field_name):
                                column_attr = getattr(model_class, field_name)
                                if isinstance(column_attr, InstrumentedAttribute):
                                    # Parse value
                                    value = self._parse_value(value_str)
                                    
                                    # Build condition based on operator
                                    if op == '==':
                                        conditions.append(column_attr == value)
                                    elif op == '!=':
                                        conditions.append(column_attr != value)
                                    elif op == '>':
                                        conditions.append(column_attr > value)
                                    elif op == '<':
                                        conditions.append(column_attr < value)
                                    elif op == '>=':
                                        conditions.append(column_attr >= value)
                                    elif op == '<=':
                                        conditions.append(column_attr <= value)
                                    elif op == ' in ':
                                        if isinstance(value, (list, tuple)):
                                            conditions.append(column_attr.in_(value))
                                        else:
                                            logger.warning(f"Value for 'in' operator must be a list or tuple: {value_str}")
                                    elif op == ' not in ':
                                        if isinstance(value, (list, tuple)):
                                            conditions.append(~column_attr.in_(value))
                                        else:
                                            logger.warning(f"Value for 'not in' operator must be a list or tuple: {value_str}")
                                    break
                        break
                        
            except Exception as e:
                logger.warning(f"Could not parse condition '{cond_str}': {e}")
        
        return conditions
    
    def _parse_value(self, value_str: str) -> Any:
        """Parse a value string to Python object"""
        value_str = value_str.strip()
        
        # Try to evaluate as Python literal
        try:
            return eval(value_str, {'__builtins__': {}})
        except:
            pass
        
        # Try to parse as string (remove quotes)
        if (value_str.startswith('"') and value_str.endswith('"')) or \
           (value_str.startswith("'") and value_str.endswith("'")):
            return value_str[1:-1]
        
        # Try common literals
        if value_str.lower() == 'true':
            return True
        if value_str.lower() == 'false':
            return False
        if value_str.lower() == 'none' or value_str.lower() == 'null':
            return None
        
        # Try to parse as number
        try:
            if '.' in value_str:
                return float(value_str)
            else:
                return int(value_str)
        except:
            pass
        
        # Return as string
        return value_str


from __future__ import annotations
import uuid
from opentelemetry.context import Context
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from service_forge.workflow.trigger import Trigger

class TriggerEvent:
    def __init__(
        self,
        trigger_node: Trigger,
        task_id: uuid.UUID,
        trace_context: Context,
        user_id: str,
    ):
        self.trigger_node = trigger_node
        self.task_id = task_id
        self.trace_context = trace_context
        self.user_id = user_id


from __future__ import annotations

import json
import requests
from abc import abstractmethod
from typing import TYPE_CHECKING, Any
from typing_extensions import override
from enum import Enum
from loguru import logger
from .workflow_event import WorkflowResult

if TYPE_CHECKING:
    from .node import Node
    from .workflow import Workflow

class CallbackEvent(Enum):
    ON_WORKFLOW_START = "on_workflow_start"
    ON_WORKFLOW_END = "on_workflow_end"
    ON_WORKFLOW_ERROR = "on_workflow_error"
    ON_NODE_START = "on_node_start"
    ON_NODE_END = "on_node_end"
    ON_NODE_OUTPUT = "on_node_output"
    ON_NODE_STREAM_OUTPUT = "on_node_stream_output"

class WorkflowCallback:
    @abstractmethod
    async def on_workflow_start(self, workflow: Workflow) -> None:
        ...

    @abstractmethod
    async def on_workflow_end(self, workflow: Workflow, output: Any) -> None:
        pass

    @abstractmethod
    async def on_workflow_error(self, workflow: Workflow, node: Node, error: Any) -> None:
        pass

    @abstractmethod
    async def on_node_start(self, node: Node) -> None:
        ...

    @abstractmethod
    async def on_node_end(self, node: Node) -> None:
        ...

    @abstractmethod
    async def on_node_stream_output(self, node: Node, output: Any) -> None:
        ...

class BuiltinWorkflowCallback(WorkflowCallback):
    def __init__(self):
        self._websocket_manager = None
    
    def _get_websocket_manager(self):
        if self._websocket_manager is None:
            from service_forge.api.routers.websocket.websocket_manager import websocket_manager
            self._websocket_manager = websocket_manager
        return self._websocket_manager
    
    def _serialize_result(self, result: Any) -> Any:
        try:
            json.dumps(result)
            return result
        except (TypeError, ValueError):
            return str(result)
    
    def _get_service_name(self) -> str | None:
        try:
            from service_forge.current_service import get_service
            service = get_service()
            if service is not None:
                return service.config.name
        except Exception:
            pass
        return None
    
    async def _send_service_event(self, workflow: Workflow, event: str) -> None:
        if workflow.entry_config is None or workflow.entry_config.url is None:
            return
        
        url = None
        try:
            service_name = self._get_service_name()
            if service_name is None:
                logger.warning("Cannot get service_name, skip sending service-event")
                return
            
            url = f"{workflow.entry_config.url.rstrip('/')}/service-event"
            payload = {
                "user_id": str(workflow.trigger_event.user_id),
                "service_name": service_name,
                "workflow_name": workflow.name,
                "task_id": str(workflow.trigger_event.task_id),
                "event": event
            }
            
            response = requests.post(url, json=payload, timeout=5)
            response.raise_for_status()
        except Exception as e:
            url_str = url if url else workflow.entry_config.url if workflow.entry_config else "unknown"
            logger.error(f"Failed to send service-event to {url_str}: {e}")
    
    @override
    async def on_workflow_start(self, workflow: Workflow) -> None:
        ...

    @override
    async def on_workflow_end(self, workflow: Workflow, output: Any) -> None:
        workflow_result = WorkflowResult(result=output, is_end=True, is_error=False)
        
        if workflow.trigger_event.task_id in workflow.real_trigger_node.result_queues:
            workflow.real_trigger_node.result_queues[workflow.trigger_event.task_id].put_nowait(workflow_result)
        if workflow.trigger_event.task_id in workflow.real_trigger_node.stream_queues:
            workflow.real_trigger_node.stream_queues[workflow.trigger_event.task_id].put_nowait(workflow_result)

        # print(workflow.entry_config, workflow.trigger_event.task_id, workflow.trigger_event.user_id)
        
        await self._send_service_event(workflow, "on_workflow_end")
        
        try:
            manager = self._get_websocket_manager()
            message = {
                "type": "workflow_end",
                "task_id": str(workflow.trigger_event.task_id),
                "result": self._serialize_result(output),
                "is_end": True,
                "is_error": False
            }
            await manager.send_to_task(workflow.trigger_event.task_id, message)
        except Exception as e:
            logger.error(f"发送 workflow_end 消息到 websocket 失败: {e}")

    @override
    async def on_workflow_error(self, workflow: Workflow, node: Node | None, error: Any) -> None:
        workflow_result = WorkflowResult(result=error, is_end=False, is_error=True)
        
        if workflow.trigger_event.task_id in workflow.real_trigger_node.result_queues:
            workflow.real_trigger_node.result_queues[workflow.trigger_event.task_id].put_nowait(workflow_result)
        if workflow.trigger_event.task_id in workflow.real_trigger_node.stream_queues:
            workflow.real_trigger_node.stream_queues[workflow.trigger_event.task_id].put_nowait(workflow_result)
        
        await self._send_service_event(workflow, "on_workflow_error")
        
        try:
            manager = self._get_websocket_manager()
            message = {
                "type": "workflow_error",
                "task_id": str(workflow.trigger_event.task_id),
                "node": node.name if node else None,
                "error": self._serialize_result(error),
                "is_end": False,
                "is_error": True
            }
            await manager.send_to_task(workflow.trigger_event.task_id, message)
        except Exception as e:
            logger.error(f"发送 workflow_error 消息到 websocket 失败: {e}")

    @override
    async def on_node_start(self, node: Node) -> None:
        try:
            manager = self._get_websocket_manager()
            message = {
                "type": "node_start",
                "task_id": str(node.workflow.trigger_event.task_id),
                "node": node.name,
                "is_end": False,
                "is_error": False
            }
            await manager.send_to_task(node.workflow.trigger_event.task_id, message)
        except Exception as e:
            logger.error(f"发送 node_start 消息到 websocket 失败: {e}")

    @override
    async def on_node_end(self, node: Node) -> None:
        try:
            manager = self._get_websocket_manager()
            message = {
                "type": "node_end",
                "task_id": str(node.workflow.trigger_event.task_id),
                "node": node.name,
                "is_end": False,
                "is_error": False
            }
            await manager.send_to_task(node.workflow.trigger_event.task_id, message)
        except Exception as e:
            logger.error(f"发送 node_end 消息到 websocket 失败: {e}")

    @override
    async def on_node_stream_output(self, node: Node, output: Any) -> None:
        workflow_result = WorkflowResult(result=output, is_end=False, is_error=False)
        
        if node.workflow.trigger_event.task_id in node.workflow.real_trigger_node.stream_queues:
            node.workflow.real_trigger_node.stream_queues[node.workflow.trigger_event.task_id].put_nowait(workflow_result)
        
        try:
            manager = self._get_websocket_manager()
            message = {
                "type": "node_stream_output",
                "task_id": str(node.workflow.trigger_event.task_id),
                "node": node.name,
                "output": self._serialize_result(output),
                "is_end": False,
                "is_error": False
            }
            await manager.send_to_task(node.workflow.trigger_event.task_id, message)
        except Exception as e:
            logger.error(f"发送 node_stream_output 消息到 websocket 失败: {e}")